// --------------------------------------------------------------------
// Calculator 5
// Parses expressions and builds expression tree
// --------------------------------------------------------------------

import scala.io.StdIn.readLine

object Calculator5 {

  def expectSymbol(tok: Tokenizer, s: String) {
    if (!tok.token.isSymbol(s))
      throw new SyntaxError(tok.tokenPos, "expected '%s'".format(s))
    tok.next()
  }
  
  // an expression is a sum of terms
  def parseExpression(tok: Tokenizer): Expression = {
    var result = parseTerm(tok)
    var t = tok.token
    while (t.isSymbol("+") || t.isSymbol("-")) {
      val op = t.text
      tok.next()
      result = Binary(op, result, parseTerm(tok))
      t = tok.token
    }
    result
  }
  
  // a term is a product of items
  def parseTerm(tok: Tokenizer): Expression = {
    var result = parseFactor(tok)
    var t = tok.token
    while (t.isSymbol("*") || t.isSymbol("/")) {
      val op = t.text
      tok.next()
      result = Binary(op, result, parseFactor(tok))
      t = tok.token
    }
    result
  }
  
  // a factor is x^y^z
  def parseFactor(tok: Tokenizer): Expression = {
    var t = tok.token
    val negate = t.isSymbol("-")
    if (t.isSymbol("-") || t.isSymbol("+"))
      tok.next()
    
    var result = parseItem(tok)
    t = tok.token
    while (t.isSymbol("^")) {
      tok.next()
      val rhs = parseFactor(tok)
      result = Binary("^", result, rhs)
      t = tok.token
    }
    if (negate)
      Unary("-", result)
    else
      result
  }
  
  // a single item: number, or expression in parentheses
  def parseItem(tok: Tokenizer): Expression = {
    val startPos = tok.tokenPos
    val t = tok.token
    tok.next()
    
    if (t.isSymbol("(")) {
      val result = parseExpression(tok)
      expectSymbol(tok, ")")
      result
    } else if (t.isNumber) {
      Number(t.number)
    } else if (t.isIdentifier) {
      Variable(t.text)
    } else
      // anything else is an error
      throw new SyntaxError(startPos, "expected number, identifier, or '('")
  }
  
  def main(args: Array[String]) {
    println("Welcome to KAIST SuperCalculator!")
    while (true) {
      val oneLine = readLine("> ")
      // terminate on End Of File 
      if (oneLine == null)
	return
      if (oneLine != "") {
	val tok = new Tokenizer(oneLine)
	try {
	  val first = tok.token
	  tok.next()
	  val second = tok.token
	  tok.reset()
	  // terminate if user says "exit"
	  if (first.isIdentifier && first.text == "exit")
	    return
	  // <identifier> = <expression> is variable assignment
	  if (first.isIdentifier && second.isSymbol("=")) {
	    tok.next()
	    tok.next()
	    val expr = parseExpression(tok)
	    if (!tok.token.isStop)
	      throw new SyntaxError(tok.tokenPos, 
				    "expected operator or end of input")
	    println("Assignment is not yet implemented")
	  } else {
	    val expr = parseExpression(tok)
	    if (!tok.token.isStop)
	      throw new SyntaxError(tok.tokenPos, 
				    "expected operator or end of input")
	    println("==> " + expr)
	  }
	} catch {
	  case e: SyntaxError => 
	    printf("Syntax error: %s\n%s\n%s\n", e.msg,
		   oneLine, " " * e.pos + "^")
	}
      }
    }
  }
}
