// --------------------------------------------------------------------
// Calculator 6
// Evaluates expressions and stores variable values
// --------------------------------------------------------------------

import scala.io.StdIn.readLine

class EvaluationError(val msg: String) extends Exception

object Calculator6 {

  var variables = Map[String, Double]()

  def evaluate(expr: Expression): Double = {
    expr match {
      case Number(v) => v
      case Variable(name) => 
	if (variables contains name)
	  variables(name)
	else
	  throw new EvaluationError("Undefined variable: " + name)
      case Binary("+", left, right) => evaluate(left) + evaluate(right)
      case Binary("-", left, right) => evaluate(left) - evaluate(right)
      case Binary("*", left, right) => evaluate(left) * evaluate(right)
      case Binary("/", left, right) => evaluate(left) / evaluate(right)
      case Binary("^", left, right) => 
	math.pow(evaluate(left), evaluate(right))
      case Unary("-", arg) => -evaluate(arg)
      case Function("sqrt", arg) => math.sqrt(evaluate(arg))
      case Function("sin", arg) => math.sin(evaluate(arg))
      case Function("cos", arg) => math.cos(evaluate(arg))
      case Function("tan", arg) => math.tan(evaluate(arg))
      case Function(name, _) => 
	throw new EvaluationError("Unknown function: " + name)
    }
  }

  def expectSymbol(tok: Tokenizer, s: String) {
    if (!tok.token.isSymbol(s))
      throw new SyntaxError(tok.tokenPos, "expected '%s'".format(s))
    tok.next()
  }
  
  // an expression is a sum of terms
  def parseExpression(tok: Tokenizer): Expression = {
    var result = parseTerm(tok)
    var t = tok.token
    while (t.isSymbol("+") || t.isSymbol("-")) {
      val op = t.text
      tok.next()
      result = Binary(op, result, parseTerm(tok))
      t = tok.token
    }
    result
  }
  
  // a term is a product of items
  def parseTerm(tok: Tokenizer): Expression = {
    var result = parseFactor(tok)
    var t = tok.token
    while (t.isSymbol("*") || t.isSymbol("/")) {
      val op = t.text
      tok.next()
      result = Binary(op, result, parseFactor(tok))
      t = tok.token
    }
    result
  }
  
  // a factor is x^y^z
  def parseFactor(tok: Tokenizer): Expression = {
    var t = tok.token
    val negate = t.isSymbol("-")
    if (t.isSymbol("-") || t.isSymbol("+"))
      tok.next()
    
    var result = parseItem(tok)
    t = tok.token
    while (t.isSymbol("^")) {
      tok.next()
      val rhs = parseFactor(tok)
      result = Binary("^", result, rhs)
      t = tok.token
    }
    if (negate)
      Unary("-", result)
    else
      result
  }
  
  // a single item: number, or expression in parentheses
  def parseItem(tok: Tokenizer): Expression = {
    val startPos = tok.tokenPos
    val t = tok.token
    tok.next()
    
    if (t.isSymbol("(")) {
      val result = parseExpression(tok)
      expectSymbol(tok, ")")
      result
    } else if (t.isNumber) {
      Number(t.number)
    } else if (t.isIdentifier) {
      if (tok.token.isSymbol("(")) {
	// function call
	val function = t.text
	tok.next()
	val arg = parseExpression(tok)
	expectSymbol(tok, ")")
	Function(function, arg)
      } else
	Variable(t.text)
    } else
      // anything else is an error
      throw new SyntaxError(startPos, "expected number, identifier, or '('")
  }
  
  def main(args: Array[String]) {
    println("Welcome to KAIST SuperCalculator!")
    while (true) {
      val oneLine = readLine("> ")
      // terminate on End Of File 
      if (oneLine == null)
	return
      if (oneLine != "") {
	val tok = new Tokenizer(oneLine)
	try {
	  val first = tok.token
	  tok.next()
	  val second = tok.token
	  tok.reset()
	  // terminate if user says "exit"
	  if (first.isIdentifier && first.text == "exit")
	    return
	  // <identifier> = <expression> is variable assignment
	  if (first.isIdentifier && second.isSymbol("=")) {
	    tok.next()
	    tok.next()
	    val expr = parseExpression(tok)
	    if (!tok.token.isStop)
	      throw new SyntaxError(tok.tokenPos, 
				    "expected operator or end of input")
	    val value = evaluate(expr)
	    variables += (first.text -> value)
	    println(first.text + " = " + value)
	  } else {
	    val expr = parseExpression(tok)
	    if (!tok.token.isStop)
	      throw new SyntaxError(tok.tokenPos, 
				    "expected operator or end of input")
	    val value = evaluate(expr)
	    println("==> " + value)
	  }
	} catch {
	  case e: SyntaxError => 
	    printf("Syntax error: %s\n%s\n%s\n", e.msg,
		   oneLine, " " * e.pos + "^")
	  case e: EvaluationError =>
	    printf("Evaluation error: %s\n", e.msg)
	}
      }
    }
  }
}
